#copyright ReportLab Europe Limited. 2000-2006
#see license.txt for license details
__version__=''' $Id: mutils.py 35543 2009-11-26 15:12:51Z rgbecker $ '''
'''Utilities for grabbing modules and examining the contents'''
import sys, os, imp, fnmatch, re, inspect, time, __main__
from types import StringType, ListType, TupleType, InstanceType
path_join=os.path.join
path_basename=os.path.basename
path_dirname=os.path.dirname
path_isfile=os.path.isfile
path_isdir=os.path.isdir
path_splitext=os.path.splitext
path_normpath=os.path.normpath
path_normcase=os.path.normcase
_tg0 = 0
_ofile = None
def importModule(mName):
	'''return a module given the import name ie reportlab.lib.utils or whatever'''
	o, e = sys.stdout, sys.stderr
	cwd = os.getcwd()
	if _ofile: t0 = time.time()
	try:
		try:
			exec 'import '+mName in locals()
			if _ofile:
				t = time.time()
				print >>_ofile, '######### import',mName,'OK %.3f %.3f' % (t-t0, t-_tg0)
			return eval(mName,locals())
		except:
			if _ofile:
				t = time.time()
				print >>_ofile, '????????? import',mName,'FAILED %.3f %.3f' % (t-t0, t-_tg0)
				import traceback
				traceback.print_exc(None,_ofile)
			return None
	finally:
		if cwd!=os.getcwd(): os.chdir(cwd)
		sys.stdout, sys.stderr = o, e

def reloadModule(m):
	try:
		if type(m) is StringType:
			if m in sys.modules.keys():
				m = reload(sys.modules[m])
			else:
				m = importModule(m)
	except:
		if _ofile:
			import traceback
			print >>_ofile, 'reload(%s) failed' % m
			traceback.print_exc(None,_ofile)
		m = None
	return m

def loadSrcModule(mPath):
	'''return a module based on the path to a source file'''
	try:
		return imp.load_source(os.path.splitext(os.path.basename(mPath))[0],mPath,open(mPath,'r'))
	except:
		return None

def normpath(x):
	return path_normpath(path_normcase(x))

def _processPackageDir(p,dn,P):
	if _ofile: print >>_ofile, 'searching package', p, 'dn=',dn
	from reportlab.lib.utils import rl_glob, rl_isfile, rl_isdir, isCompactDistro
	if isCompactDistro():
		ext = '.pyc'
		init = '__init__'+ext
		FN = [normpath(x) for x in rl_glob(path_join(dn,'*'))]
		D = []
		dn = normpath(dn)
		for x in FN:
			x = path_dirname(x)
			if x not in D and rl_isdir(x) and path_dirname(x)==dn and rl_isfile(path_join(x,init)): D.append(x)
		F = [x for x in FN if x.endswith(ext) and path_dirname(x)==dn and not path_basename(x).startswith('__init__.')]
	else:
		ext = '.py'
		init = '__init__'+ext
		FN = [path_join(dn,x) for x in os.listdir(dn)]
		D = [x for x in FN if path_isdir(x) and path_isfile(path_join(x,init))]
		F = [x for x in FN if x.endswith(ext) and not path_basename(x).startswith('__init__.')]
	for f in F:
		mn = p+('.'+path_splitext(path_basename(f))[0])
		if mn not in P:
			if _ofile: print >>_ofile, 'appending 1',mn
			P.append(mn)
	for f in D:
		mn = p+('.'+path_basename(f))
		if mn not in P:
			if _ofile: print >>_ofile, 'appending 2',mn
			P.append(mn)

def _checkPmw():
	'''remove the Pmw loader interference'''
	mods = {}
	for n in '_Pmw', 'Pmw':
		try:
			m = sys.modules[n]
			del sys.modules[n]
			mods[n] = m
		except:
			pass
	return mods

def _uncheckPmw(mods):
	'''restore the Pmw loader interference'''
	sys.modules.update(mods)

def getModules(P=['reportlab.graphics'],X=[], MF=None, MD=None):
	'''Locate all modules descendant from elements of P'''

	if type(X) is StringType: X=[X]
	xp = ''
	for p in X:
		xp += '|'+ fnmatch.translate(p)[:-1]
	if xp!='':
		xp = re.compile(xp[1:]+'$')
	M = []
	P = type(P) is StringType and [P] or P[:]
	if MF is None: MF = []
	if MD is None: MD = []
	while P:
		p = P[0]
		del P[0]
		if xp and xp.match(p):
			if _ofile:
				t = time.time()
				print >>_ofile, '????????? import',p,'EXCLUDED %.3f %.3f' % (0, t-_tg0)
			continue
		if p=='.':
			_processPackageDir('',os.getcwd(),P)
		else:
			if _ofile: print >>_ofile, 'trying importModule(%s)' % p
			m = importModule(p)
			if m and m not in M:
				mn = m.__file__
				if _ofile: print >>_ofile, 'module %s found __file__=%s' % (m,mn)
				if mn not in MF:
					M.append(m)
					MF.append(mn)
					if _ofile: print >>_ofile, 'appending 3', m, mn
					if os.path.basename(mn)[:9]=='__init__.':
						MD.append(os.path.dirname(mn))
						#it's a package so recur
						_processPackageDir(p,os.path.dirname(mn),P)
	return M

def getModuleClassMethod(moduleName,className,methodName):
	M = importModule(moduleName)
	if not M: return None
	C = getattr(M,className,None)
	if not C: return None
	return getattr(C,methodName,None)

def getModuleClassMethodDefaults(moduleName,className,methodName,**kw):
	method = getModuleClassMethod(moduleName,className,methodName)
	if method:
		argspec = inspect.getargspec(method.im_func)
		V = argspec[3]
		N = argspec[0][1:]
		for k in kw.keys():
			i = N.index(k)
			kw[k] = (i>=0 and i<len(V)) and V[i] or None
	return kw

def showAtThisMode(obj,k,usage='basic'):
	"""returns True if this attribute should be used at this mode"""
	if usage=='private': 
		return True # at this mode show everything
	validModes = ['basic']
	if usage == 'advance':
		validModes.append('advance')
	if getAttrUsageMode(obj,k) in validModes:
		return True
	else:
		return False

def getAttributes(obj,private=0,advanced=0):
	from reportlab.graphics.shapes import Group
	from reportlab.graphics.widgetbase import PropHolder, TypedPropertyCollection
	A = []
	P = []
	isG = isinstance(obj,Group)
	if hasattr(obj,'contents') and isG:
		contents = obj.contents
		P.append(contents)
	else:
		contents = []

	def addAP(k,v,A=A,P=P):
		if (k,v) in A or v in P: return
		if type(v) is InstanceType:
				A.append((k,v))
				P.append(v)
		else:
			A.append((k,v))

	attrMap = getattr(obj,'_attrMap',None)
	R = []
	if attrMap:
		for k,a in attrMap.iteritems():
			advancedUsage = getattr(a,'_advancedUsage',0)
			if (not k.startswith('_') or private) and (hasattr(obj,k) and (not advancedUsage or advanced)):
				v = getattr(obj,k,None)
				addAP(k,v)
			else:
				R.append(k)

	for k,v in getattr(obj,'__dict__',{}).items():
		if k not in R and (not k.startswith('_') or private):
			addAP(k,v)
		else:
			R.append(k)
	g = getattr(obj,'getProperties',None)
	#pdb.set_trace()
	if g:
		for k,v in g(recur=0).items():
			if k not in R and (not k.startswith('_') or private):
				addAP(k,v)
			else:
				R.append(k)

	for c in contents:
		addAP('contents[%d]'% contents.index(c),c)

	A.sort()
	return A

def pGetClassModule(v):
	try:
		return sys.modules.get(v.__module__)
	except:
		return None

def _normpath(p):
	return os.path.normcase(os.path.normpath(p))

def _cmpMRCN(a,b):
	'''compare two named modules by refcount and then name'''
	r = cmp(sys.getrefcount(sys.modules[b]), sys.getrefcount(sys.modules[a]))
	if not r: r = cmp(b,a)
	return r

absfile=lambda x: os.path.normcase(os.path.abspath(x))

def _getmembers(m,_gm=inspect.getmembers,_ff=lambda v: v[0][:2]!='__'):
	return filter(_ff,_gm(m))

def _getfunctions(members,m):
	mfile = m.__file__
	if mfile[-1] in ['c','o']: mfile = mfile[:-1]
	try:
		a = m.__loader__.archive + os.sep
		if mfile.startswith(a):
			mfile = mfile[len(a):]
		ff = lambda v,mfile=mfile: inspect.isfunction(v[1]) and v[1].func_code.co_filename==mfile
	except:
		mfile = absfile(mfile)
		ff = lambda v,mfile=mfile: inspect.isfunction(v[1]) and absfile(v[1].func_code.co_filename)==mfile
	return filter(ff,members)

def _getclasses(members,m):
	return filter(lambda v,m=m: inspect.isclass(v[1]) and pGetClassModule(v[1])==m,members)

def _getvars(members, _ff = lambda v: not (inspect.isclass(v[1]) or inspect.ismethod(v[1])\
					or inspect.ismodule(v[1]) or inspect.isroutine(v[1]) or inspect.istraceback(v[1])\
					or inspect.iscode(v[1]) or inspect.isframe(v[1]))):
	return filter(_ff,members)

class KnownNames:
	def __init__(self,SX=([],[])):
		self._ga = {'modules':[],'classes':{},'funcs':{},'varsN':{}, 'varsI':{}, 'search':()}
		self._MD = []
		self.search = SX

	def __setattr__(self,name,value):
		if name=='search':
			self._search(value)
		elif name[0]=='_':
			self.__dict__[name] = value
		else:
			raise AttributeError, name

	def __getattr__(self,name):
		if self._ga.has_key(name): return self._ga[name]
		raise AttributeError, name

	def findName(self,name):
		R = {}
		A = self._ga
		for k,v in A.items():
			if k=='search': continue
			if k == 'classObjs':
				if name in v: R['module'] = v[name]
			elif k == 'modules':
				N = map(lambda m: m.__name__,v)
				if name in N: R[''] = name
			else:
				if	v.has_key(name): R[k] = v[name]
		return R

	def checkPath(self,path):
		'''return 1 if path is covered by our modules/dirs else 0'''
		D = _normpath(os.path.dirname(path))
		for d in self._MD:
			if D==_normpath(d): return 1
		return 0

	def reloadModule(self,M):
		'''reload a single module or set of modules R'''
		if type(M) not in (TupleType,ListType): R = [M]
		else:
			R = list(M)[:]
			R.sort()
			R.reverse()
		OM = []
		K = []
		M = R[:]
		while M:
			RR = []
			for m in M:
				om = self._delModule(m)
				if om:
					RR.append(m)
					OM.append(om[1])
					if om[0]: K.append(m)
			if not RR: break
			map(M.remove,RR)
		if not OM: return
		#if _ofile:
		#	print >>_ofile, '+++++++++++++++Final module links'
		#	for m in OM:
		#		print >>_ofile, m.__name__, sys.getrefcount(m)
		del OM
		#if _ofile and M: print >>_ofile, '#####UNDELETED modules',M
		ga = self._ga
		M = ga['modules']
		#if _ofile and M: print >>_ofile, '#####UNDELETED known modules',M
		C = ga['classes']
		F = ga['funcs']
		VN = ga['varsN']
		VI = ga['varsI']
		R.reverse()
		while R:
			RR = []
			for m in R:
				im = importModule(m)
				if im: RR.append(m)
				if m in K:
					self._addModule(m,M,C,F,VN,VI)
					self._finish(M,C,F,VN,VI)
			if not RR: break
			map(R.remove,RR)

	def _search(self,SX):
		ga=self._ga
		if type(SX) not in (ListType,TupleType): SX = [SX,ga['search'][1]]
		if SX==ga['search']: return
		ga['search'] = SX
		self._MD = []
		M = getModules(SX[0],SX[1],MD=self._MD)
		if _ofile:
			t = time.time()
			print >>_ofile, "	getmodules done %.3f" % (t-_tg0)
		if M==ga['modules']: return
		F={}
		C={}
		VN={}
		VI={}
		for m in M:
			self._addModule(m,M,C,F,VN,VI)
		if _ofile:
			t = time.time()
			print >>_ofile, "	_addModule done %.3f" % (t-_tg0)
		self._finish(M,C,F,VN,VI)
		if _ofile:
			t = time.time()
			print >>_ofile, "	_finish done %.3f" % (t-_tg0)


	def _addModule(self,m,M,C,F,VN,VI):
		if type(m) is StringType:
			m = importModule(m)
			if m is None: return
		if m not in M: M.append(m)
		members = _getmembers(m)
		for X,V in ((F,_getfunctions(members,m)),(C,_getclasses(members,m))):
			for v in V:
				n = v[0]
				if not X.has_key(n): X[n] = [m]
				else: X[n].append(m)
		V = _getvars(members)
		for v in V:
			i = id(v[1])
			n = v[0]
			if not VN.has_key(n): VN[n] = [m]
			else: VN[n].append(m)
			n = (n,m)
			if not VI.has_key(i): VI[i] = [n]
			else: VI[i].append(n)

	def _delModule(self,m):
		'''delete a module by name and remove it from sys.modules
		return (flag,module) flag is 1 if the module was in the controlled set
		'''
		if type(m) is StringType:
			m = sys.modules.get(m,None)
			if m is None: return None
		del sys.modules[m.__name__]
		ga = self._ga
		M = ga['modules']
		r = m in M
		if r:
			C = ga['classes']
			F = ga['funcs']
			VN = ga['varsN']
			VI = ga['varsI']
			members = filter(lambda v: v[0][:2]!='__',inspect.getmembers(m))
			for X,V in ((F,_getfunctions(members,m)),(C,_getclasses(members,m))):
				for v in V:
					n = v[0]
					try:
						X[n].remove(m)
						if X[n]==[]: del X[n]
					except:
						pass

			V = _getvars(members)
			for v in V:
				i = id(v[1])
				n = v[0]
				try:
					VN[n].remove(m)
					if VN[n]==[]: del VN[n]
				except:
					pass
				n = (n,m)
				try:
					VI[i].remove(n)
					if VN[i]==[]: del VN[i]
				except:
					pass
			M.remove(m)
			self._finish(M,C,F,VN,VI)
		#if _ofile: print >>_ofile, 'delModule(%s) rc=%d known=%d len(M)==%d' % (m.__name__,sys.getrefcount(m),r,len(M))
		return r,m


	def _finish(self,M,C,F,VN,VI):
		ga=self._ga
		ga['modules'] = M
		ga['classes'] = C
		ga['funcs'] = F
		ga['varsN'] = VN
		ga['varsI'] = VI
		CO = []
		for n,M in C.items():
			for m in M:
				CO.append(getattr(m,n))
		ga['classObjs'] = CO

def _getKnown(gM=0,gK=1,dbg=0,
				search = ['fidelity.epscharts','reportlab.lib','reportlab','rlextra'],
				exclude = ['*fidelity.epscharts.do_*','reportlab.demo.*','reportlab.test*',
							'reportlab.doc*','reportlab.utils.*','rlextra.preppy.flintstones*',
							'*fidelity.epscharts.updatedoc','reportlab.lib.graphdocpy',
							'*.guiedit.guiedit','*.guidialogs','*.CVS',
							],
			):
	global _tg0, _ofile
	try:
		from rlextra.graphics.guiedit.guiedit import _getConfig
		search = config.package.search
		config = _getConfig(os.path.join(os.path.dirname(sys.argv[0]),'_guiedit.ini'))
		exclude = config.package.exclude+['*.guiedit.guiedit','*.guidialogs']
	except:
		pass
	_ofile=dbg and sys.stdout or None
	if gM:
		_tg0 = time.time()
		pmwMods = _checkPmw()
		M = getModules(search,exclude)
		_uncheckPmw(pmwMods)
		print >>_ofile, 'getModules %.2f' % (time.time()-_tg0)
	if gK:
		_tg0 = time.time()
		pmwMods = _checkPmw()
		globals()['_known']=K=KnownNames((search,exclude))
		_uncheckPmw(pmwMods)
		print >>_ofile, 'KnownNames %.2f' % (time.time()-_tg0)
		print >>_ofile, '#M=%d #F=%d #C=%d #V=%d'%(len(K.modules),len(K.funcs.keys()),
					len(K.classes.keys()),len(K.varsN.keys()))

class AssignCheckerLine:
	def __init__(self,text):
		self.text = text
		T = self.tokens(text)
		s = 0
		self.deletions = []
		self.assigns = []
		a = self.assigns.append
		for i in (i for i,x in enumerate(T) if x[1].endswith('=') and (i<2 or T[i-2][1] not in (',','('))):
			lhs = []
			for j in xrange(s,i):
				t = T[j][1]
				if t[-1] in "\"'":
					t = repr(eval(t))
				lhs.append(t)
			a((''.join(lhs),T[s][2][1],T[i+1][2][1]))
			s = i+1
		self.nAssigns = len(self.assigns)

	def delete(self,j):
		self.deletions.append(j)

	def restore(self,j):
		self.deletions.remove(j)

	def render(self):
		if len(self.deletions)==self.nAssigns: return ''
		t = self.text
		a = self.assigns
		for j in self.deletions:
			s,f = a[j][1:]
			t = t[:s]+(' '*(f-s))+t[f:]
		return t.lstrip()

	def LHS(self):
		L = getattr(self,'_LHS',None)
		if L==None:
			L = self._LHS = list(a[0] for a in self.assigns)
		return L
	LHS = property(LHS)

	@staticmethod
	def tokens(s):
		import tokenize, StringIO
		return list(tokenize.generate_tokens(StringIO.StringIO(s).readline))

class AssignChecker:
	def __init__(self,lines):
		self.lines = [AssignCheckerLine(l) for l in lines]

	def repeatedAssigns(self):
		RA = getattr(self,'_repeatedAssigns',None)
		if RA is None:
			LHS = {}
			for i,l in enumerate(self.lines):
				for j,lhs in enumerate(l.LHS):
					LHS.setdefault(lhs,[]).append((i,j))
			RA = self._repeatedAssigns = []
			for v in LHS.itervalues():
				if len(v)>1:
					RA.extend(v[:-1])
			RA.sort()
		return RA
	repeatedAssigns = property(repeatedAssigns)

if __name__=='__main__':
	if '-prof' in sys.argv:
		import profile
		profile.run('_getKnown()','mutils.prof')
	else:
		_getKnown(dbg=1)
