#!/usr/bin/env python
"""Removes all markup from a string

>>> 2+2
4
>>> stripTags('hello')
'hello'
>>> stripTags('hello<a href="crap"> world</a>')
'hello world'
>>> stripTags('goodbye<!--stuff--> comments')
'goodbye comments'
>>> stripTags('Entities &#amp; stuff')
'Entities &#amp; stuff'
>>> stripTags('Fr&#233;gate Island')   #accents etc preserved
'Fr&#233;gate Island'
>>> stripTags('Space&#nbsp;here')   #entities preserved
'Space&#nbsp;here'
>>> resolveEntities('Fr&#233;gate Island')   #accents etc preserved
'Fr\\xc3\\xa9gate Island'
"""
import sgmllib
from htmlentitydefs import name2codepoint
from rlextra.radxml.xmlutils import nakedAmpFix

def stripTags(text):
    return TagStripper().strip(text)

def resolveEntities(text):
    "Turn them to UTF8 text"
    return EntityResolver().resolve(text)

class TagStripper(sgmllib.SGMLParser):
    def strip(self, text):
        self._output = []
        self.reset()
        self.feed(nakedAmpFix(text).replace('<br/>', '<br />'))
        return ''.join(self._output)
    
    def handle_data(self, data):
        self._output.append(data)

    def convert_entityref(self, ref):
        return unichr(name2codepoint.get(ref, 0))

    def convert_charref(self, ref):
        "We want char refs to be preserved"
        #print 'ref=',ref, type(ref)
        return '&#%s;' % ref
    

class EntityResolver(sgmllib.SGMLParser):
    def resolve(self, text):
        self._output = []
        self.reset()
        self.feed(text)
        return ''.join(self._output)
    
    def handle_data(self, data):
        self._output.append(data)

    def convert_charref(self, ref):
        ch = sgmllib.SGMLParser.convert_charref(self, ref)
        if ch:
            return ch.decode('latin-1').encode('utf8')
    

if __name__=='__main__':
    import doctest, tagstripper
    doctest.testmod(tagstripper)

